@@ -33,6 +33,8 @@ static GMainLoop *glib_loop;
 static qb_array_t *gio_map;
 #endif /* HAVE_GLIB */
 
+#define ONE_MEG 1048576
+
 static int32_t use_glib = QB_FALSE;
 static int32_t use_events = QB_FALSE;
 static qb_loop_t *bms_loop;
@@ -181,8 +183,9 @@ show_usage(const char *name)
 #ifdef HAVE_GLIB
 struct gio_to_qb_poll {
 	gboolean is_used;
-	GIOChannel *channel;
 	int32_t events;
+	int32_t source;
+	int32_t fd;
 	void *data;
 	qb_ipcs_dispatch_fn_t fn;
 	enum qb_loop_priority p;
@@ -197,6 +200,16 @@ gio_read_socket(GIOChannel * gio, GIOCondition condition, gpointer data)
 	return (adaptor->fn(fd, condition, adaptor->data) == 0);
 }
 
+static void
+gio_poll_destroy(gpointer data)
+{
+	struct gio_to_qb_poll *adaptor = (struct gio_to_qb_poll *)data;
+
+	qb_log(LOG_DEBUG, "fd %d adaptor destroyed\n", adaptor->fd);
+	adaptor->is_used = QB_FALSE;
+	adaptor->fd = 0;
+}
+
 static int32_t
 my_g_dispatch_add(enum qb_loop_priority p, int32_t fd, int32_t evts,
 		  void *data, qb_ipcs_dispatch_fn_t fn)
@@ -218,14 +231,19 @@ my_g_dispatch_add(enum qb_loop_priority p, int32_t fd, int32_t evts,
 		return -ENOMEM;
 	}
 
-	adaptor->channel = channel;
 	adaptor->fn = fn;
 	adaptor->events = evts;
 	adaptor->data = data;
 	adaptor->p = p;
 	adaptor->is_used = TRUE;
+	adaptor->fd = fd;
+
+	adaptor->source = g_io_add_watch_full(channel, G_PRIORITY_DEFAULT, evts, gio_read_socket, adaptor, gio_poll_destroy);
+
+	/* we are handing the channel off to be managed by mainloop now.
+	 * remove our reference. */
+	g_io_channel_unref(channel);
 
-	g_io_add_watch(channel, evts, gio_read_socket, adaptor);
 	return 0;
 }
 
@@ -241,7 +259,7 @@ my_g_dispatch_del(int32_t fd)
 {
 	struct gio_to_qb_poll *adaptor;
 	if (qb_array_index(gio_map, fd, (void **)&adaptor) == 0) {
-		g_io_channel_unref(adaptor->channel);
+		g_source_remove(adaptor->source);
 		adaptor->is_used = FALSE;
 	}
 	return 0;
@@ -279,6 +297,7 @@ main(int32_t argc, char *argv[])
 {
 	const char *options = "mpseugh";
 	int32_t opt;
+	int32_t rc;
 	enum qb_ipc_type ipc_type = QB_IPC_NATIVE;
 	struct qb_ipcs_service_handlers sh = {
 		.connection_accept = s1_connection_accept_fn,
@@ -336,17 +355,30 @@ main(int32_t argc, char *argv[])
 		qb_perror(LOG_ERR, "qb_ipcs_create");
 		exit(1);
 	}
+	/* This forces the clients to use a minimum buffer size */
+	qb_ipcs_enforce_buffer_size(s1, ONE_MEG);
+
 	if (!use_glib) {
 		bms_loop = qb_loop_create();
 		qb_ipcs_poll_handlers_set(s1, &ph);
-		qb_ipcs_run(s1);
+		rc = qb_ipcs_run(s1);
+		if (rc != 0) {
+			errno = -rc;
+			qb_perror(LOG_ERR, "qb_ipcs_run");
+			exit(1);
+		}
 		qb_loop_run(bms_loop);
 	} else {
 #ifdef HAVE_GLIB
 		glib_loop = g_main_loop_new(NULL, FALSE);
 		gio_map = qb_array_create_2(16, sizeof(struct gio_to_qb_poll), 1);
 		qb_ipcs_poll_handlers_set(s1, &glib_ph);
-		qb_ipcs_run(s1);
+		rc = qb_ipcs_run(s1);
+		if (rc != 0) {
+			errno = -rc;
+			qb_perror(LOG_ERR, "qb_ipcs_run");
+			exit(1);
+		}
 		g_main_loop_run(glib_loop);
 #else
 		qb_log(LOG_ERR,
