@@ -57,7 +57,9 @@ qb_ipcs_create(const char *name,
 	s->pid = getpid();
 	s->needs_sock_for_poll = QB_FALSE;
 	s->poll_priority = QB_LOOP_MED;
-	s->ref_count = 1;
+
+	/* Initial alloc ref */
+	qb_ipcs_ref(s);
 
 	s->service_id = service_id;
 	(void)strlcpy(s->name, name, NAME_MAX);
@@ -85,6 +87,19 @@ qb_ipcs_poll_handlers_set(struct qb_ipcs_service *s,
 	s->poll_fns.dispatch_del = handlers->dispatch_del;
 }
 
+void
+qb_ipcs_service_context_set(qb_ipcs_service_t* s,
+			    void *context)
+{
+	s->context = context;
+}
+
+void *
+qb_ipcs_service_context_get(qb_ipcs_service_t* s)
+{
+	return s->context;
+}
+
 int32_t
 qb_ipcs_run(struct qb_ipcs_service *s)
 {
@@ -93,7 +108,9 @@ qb_ipcs_run(struct qb_ipcs_service *s)
 	if (s->poll_fns.dispatch_add == NULL ||
 	    s->poll_fns.dispatch_mod == NULL ||
 	    s->poll_fns.dispatch_del == NULL) {
-		return -EINVAL;
+
+		res = -EINVAL;
+		goto run_cleanup;
 	}
 
 	switch (s->type) {
@@ -115,15 +132,19 @@ qb_ipcs_run(struct qb_ipcs_service *s)
 		res = -EINVAL;
 		break;
 	}
-	if (res < 0) {
-		qb_ipcs_unref(s);
-		return res;
+
+	if (res == 0) {
+		res = qb_ipcs_us_publish(s);
+		if (res < 0) {
+			(void)qb_ipcs_us_withdraw(s);
+			goto run_cleanup;
+		}
 	}
-	res = qb_ipcs_us_publish(s);
+
+run_cleanup:
 	if (res < 0) {
-		(void)qb_ipcs_us_withdraw(s);
+		/* Failed to run services, removing initial alloc reference. */
 		qb_ipcs_unref(s);
-		return res;
 	}
 
 	return res;
@@ -172,8 +193,7 @@ qb_ipcs_request_rate_limit(struct qb_ipcs_service *s,
 		break;
 	}
 
-	for (pos = s->connections.next, n = pos->next;
-	     pos != &s->connections; pos = n, n = pos->next) {
+	qb_list_for_each_safe(pos, n, &s->connections) {
 
 		c = qb_list_entry(pos, struct qb_ipcs_connection, list);
 		qb_ipcs_connection_ref(c);
@@ -202,23 +222,11 @@ void
 qb_ipcs_unref(struct qb_ipcs_service *s)
 {
 	int32_t free_it;
-	struct qb_ipcs_connection *c = NULL;
-	struct qb_list_head *pos;
-	struct qb_list_head *n;
 
 	assert(s->ref_count > 0);
 	free_it = qb_atomic_int_dec_and_test(&s->ref_count);
 	if (free_it) {
 		qb_util_log(LOG_DEBUG, "%s() - destroying", __func__);
-		for (pos = s->connections.next, n = pos->next;
-		     pos != &s->connections; pos = n, n = pos->next) {
-			c = qb_list_entry(pos, struct qb_ipcs_connection, list);
-			if (c == NULL) {
-				continue;
-			}
-			qb_ipcs_disconnect(c);
-		}
-		(void)qb_ipcs_us_withdraw(s);
 		free(s);
 	}
 }
@@ -226,6 +234,23 @@ qb_ipcs_unref(struct qb_ipcs_service *s)
 void
 qb_ipcs_destroy(struct qb_ipcs_service *s)
 {
+	struct qb_ipcs_connection *c = NULL;
+	struct qb_list_head *pos;
+	struct qb_list_head *n;
+
+	if (s == NULL) {
+		return;
+	}
+	qb_list_for_each_safe(pos, n, &s->connections) {
+		c = qb_list_entry(pos, struct qb_ipcs_connection, list);
+		if (c == NULL) {
+			continue;
+		}
+		qb_ipcs_disconnect(c);
+	}
+	(void)qb_ipcs_us_withdraw(s);
+
+	/* service destroyed, remove initial alloc ref */
 	qb_ipcs_unref(s);
 }
 
@@ -272,7 +297,7 @@ qb_ipcs_response_send(struct qb_ipcs_connection *c, const void *data,
 	} else if (res == -EAGAIN || res == -ETIMEDOUT) {
 		struct qb_ipc_one_way *ow = _response_sock_one_way_get(c);
 		if (ow) {
-			ssize_t res2  = qb_ipc_us_ready(ow, 0, POLLOUT);
+			ssize_t res2 = qb_ipc_us_ready(ow, &c->setup, 0, POLLOUT);
 			if (res2 < 0) {
 				res = res2;
 			}
@@ -300,7 +325,7 @@ qb_ipcs_response_sendv(struct qb_ipcs_connection * c, const struct iovec * iov,
 	} else if (res == -EAGAIN || res == -ETIMEDOUT) {
 		struct qb_ipc_one_way *ow = _response_sock_one_way_get(c);
 		if (ow) {
-			ssize_t res2  = qb_ipc_us_ready(ow, 0, POLLOUT);
+			ssize_t res2 = qb_ipc_us_ready(ow, &c->setup, 0, POLLOUT);
 			if (res2 < 0) {
 				res = res2;
 			}
@@ -317,6 +342,10 @@ resend_event_notifications(struct qb_ipcs_connection *c)
 {
 	ssize_t res = 0;
 
+	if (!c->service->needs_sock_for_poll) {
+		return res;
+	}
+
 	if (c->outstanding_notifiers > 0) {
 		res = qb_ipc_us_send(&c->setup, c->receive_buf,
 				     c->outstanding_notifiers);
@@ -324,6 +353,7 @@ resend_event_notifications(struct qb_ipcs_connection *c)
 	if (res > 0) {
 		c->outstanding_notifiers -= res;
 	}
+
 	assert(c->outstanding_notifiers >= 0);
 	if (c->outstanding_notifiers == 0) {
 		c->poll_events = POLLIN | POLLPRI | POLLNVAL;
@@ -344,6 +374,7 @@ new_event_notification(struct qb_ipcs_connection * c)
 	assert(c->outstanding_notifiers >= 0);
 	if (c->outstanding_notifiers > 0) {
 		c->outstanding_notifiers++;
+		res = resend_event_notifications(c);
 	} else {
 		res = qb_ipc_us_send(&c->setup, &c->outstanding_notifiers, 1);
 		if (res == -EAGAIN) {
@@ -366,14 +397,16 @@ qb_ipcs_event_send(struct qb_ipcs_connection * c, const void *data, size_t size)
 
 	if (c == NULL) {
 		return -EINVAL;
+	} else if (size > c->event.max_msg_size) {
+		return -EMSGSIZE;
 	}
-	qb_ipcs_connection_ref(c);
 
+	qb_ipcs_connection_ref(c);
 	res = c->service->funcs.send(&c->event, data, size);
 	if (res == size) {
 		c->stats.events++;
 		resn = new_event_notification(c);
-		if (resn < 0 && resn != -EAGAIN) {
+		if (resn < 0 && resn != -EAGAIN && resn != -ENOBUFS) {
 			errno = -resn;
 			qb_util_perror(LOG_WARNING,
 				       "new_event_notification (%s)",
@@ -382,8 +415,12 @@ qb_ipcs_event_send(struct qb_ipcs_connection * c, const void *data, size_t size)
 		}
 	} else if (res == -EAGAIN || res == -ETIMEDOUT) {
 		struct qb_ipc_one_way *ow = _event_sock_one_way_get(c);
+
+		if (c->outstanding_notifiers > 0) {
+			resn = resend_event_notifications(c);
+		}
 		if (ow) {
-			resn  = qb_ipc_us_ready(ow, 0, POLLOUT);
+			resn = qb_ipc_us_ready(ow, &c->setup, 0, POLLOUT);
 			if (resn < 0) {
 				res = resn;
 			}
@@ -420,8 +457,12 @@ qb_ipcs_event_sendv(struct qb_ipcs_connection * c,
 		}
 	} else if (res == -EAGAIN || res == -ETIMEDOUT) {
 		struct qb_ipc_one_way *ow = _event_sock_one_way_get(c);
+
+		if (c->outstanding_notifiers > 0) {
+			resn = resend_event_notifications(c);
+		}
 		if (ow) {
-			resn  = qb_ipc_us_ready(ow, 0, POLLOUT);
+			resn = qb_ipc_us_ready(ow, &c->setup, 0, POLLOUT);
 			if (resn < 0) {
 				res = resn;
 			}
@@ -437,14 +478,13 @@ qb_ipcs_connection_t *
 qb_ipcs_connection_first_get(struct qb_ipcs_service * s)
 {
 	struct qb_ipcs_connection *c;
-	struct qb_list_head *iter;
 
 	if (qb_list_empty(&s->connections)) {
 		return NULL;
 	}
-	iter = s->connections.next;
 
-	c = qb_list_entry(iter, struct qb_ipcs_connection, list);
+	c = qb_list_first_entry(&s->connections, struct qb_ipcs_connection,
+				list);
 	qb_ipcs_connection_ref(c);
 
 	return c;
@@ -455,14 +495,14 @@ qb_ipcs_connection_next_get(struct qb_ipcs_service * s,
 			    struct qb_ipcs_connection * current)
 {
 	struct qb_ipcs_connection *c;
-	struct qb_list_head *iter;
 
-	if (current == NULL || current->list.next == &s->connections) {
+	if (current == NULL ||
+	    qb_list_is_last(&current->list, &s->connections)) {
 		return NULL;
 	}
-	iter = current->list.next;
 
-	c = qb_list_entry(iter, struct qb_ipcs_connection, list);
+	c = qb_list_first_entry(&current->list, struct qb_ipcs_connection,
+				list);
 	qb_ipcs_connection_ref(c);
 
 	return c;
@@ -487,12 +527,9 @@ qb_ipcs_connection_alloc(struct qb_ipcs_service *s)
 		return NULL;
 	}
 
-	c->refcount = 1;
-	c->service = s;
 	c->pid = 0;
 	c->euid = -1;
 	c->egid = -1;
-	qb_list_init(&c->list);
 	c->receive_buf = NULL;
 	c->context = NULL;
 	c->fc_enabled = QB_FALSE;
@@ -505,6 +542,18 @@ qb_ipcs_connection_alloc(struct qb_ipcs_service *s)
 	c->event.type = s->type;
 	(void)strlcpy(c->description, "not set yet", CONNECTION_DESCRIPTION);
 
+	/* initial alloc ref */
+	qb_ipcs_connection_ref(c);
+
+	/*
+	 * The connection makes use of the service object. Give the connection
+	 * a reference to the service so we know the service can never be destroyed
+	 * until the connection is done with it.
+	 */
+	qb_ipcs_ref(s);
+	c->service = s;
+	qb_list_init(&c->list);
+
 	return c;
 }
 
@@ -526,8 +575,7 @@ qb_ipcs_connection_unref(struct qb_ipcs_connection *c)
 	}
 	if (c->refcount < 1) {
 		qb_util_log(LOG_ERR, "ref:%d state:%d (%s)",
-			    c->refcount, c->state,
-			    c->description);
+			    c->refcount, c->state, c->description);
 		assert(0);
 	}
 	free_it = qb_atomic_int_dec_and_test(&c->refcount);
@@ -537,6 +585,8 @@ qb_ipcs_connection_unref(struct qb_ipcs_connection *c)
 			c->service->serv_fns.connection_destroyed(c);
 		}
 		c->service->funcs.disconnect(c);
+		/* Let go of the connection's reference to the service */
+		qb_ipcs_unref(c->service);
 		free(c->receive_buf);
 		free(c);
 	}
@@ -555,41 +605,47 @@ qb_ipcs_disconnect(struct qb_ipcs_connection *c)
 		    __func__, c->description, c->state);
 
 	if (c->state == QB_IPCS_CONNECTION_ACTIVE) {
+		c->service->funcs.disconnect(c);
 		c->state = QB_IPCS_CONNECTION_INACTIVE;
 		c->service->stats.closed_connections++;
 
-		qb_ipcs_sockets_disconnect(c);
+		/* This removes the initial alloc ref */
+		qb_ipcs_connection_unref(c);
+
 		/* return early as it's an incomplete connection.
 		 */
 		return;
 	}
 	if (c->state == QB_IPCS_CONNECTION_ESTABLISHED) {
+		c->service->funcs.disconnect(c);
 		c->state = QB_IPCS_CONNECTION_SHUTTING_DOWN;
 		c->service->stats.active_connections--;
 		c->service->stats.closed_connections++;
-
-		qb_ipcs_sockets_disconnect(c);
 	}
 	if (c->state == QB_IPCS_CONNECTION_SHUTTING_DOWN) {
+		int scheduled_retry = 0;
 		res = 0;
 		if (c->service->serv_fns.connection_closed) {
 			res = c->service->serv_fns.connection_closed(c);
 		}
-		if (res == 0) {
-			qb_ipcs_connection_unref(c);
-		} else {
+		if (res != 0) {
 			/* OK, so they want the connection_closed
 			 * function re-run */
 			rerun_job =
 			    (qb_loop_job_dispatch_fn) qb_ipcs_disconnect;
 			res = c->service->poll_fns.job_add(QB_LOOP_LOW,
-							   c,
-							   rerun_job);
-			if (res != 0) {
-				/* last ditch attempt to cleanup */
-				qb_ipcs_connection_unref(c);
+							   c, rerun_job);
+			if (res == 0) {
+				/* this function is going to be called again.
+				 * so hold off on the unref */
+				scheduled_retry = 1;
 			}
 		}
+
+		if (scheduled_retry == 0) {
+			/* This removes the initial alloc ref */
+			qb_ipcs_connection_unref(c);
+		}
 	}
 }
 
@@ -628,7 +684,7 @@ _process_request_(struct qb_ipcs_connection *c, int32_t ms_timeout)
 	}
 	if (size < 0) {
 		if (size != -EAGAIN && size != -ETIMEDOUT) {
-			qb_util_perror(LOG_ERR,
+			qb_util_perror(LOG_DEBUG,
 				       "recv from client connection failed (%s)",
 				       c->description);
 		} else {
@@ -639,9 +695,8 @@ _process_request_(struct qb_ipcs_connection *c, int32_t ms_timeout)
 	} else if (size == 0 || hdr->id == QB_IPC_MSG_DISCONNECT) {
 		qb_util_log(LOG_DEBUG, "client requesting a disconnect (%s)",
 			    c->description);
-		qb_ipcs_disconnect(c);
-		c = NULL;
 		res = -ESHUTDOWN;
+		goto cleanup;
 	} else {
 		c->stats.requests++;
 		res = c->service->serv_fns.msg_process(c, hdr, hdr->size);
@@ -665,17 +720,6 @@ cleanup:
 #define IPC_REQUEST_TIMEOUT 10
 #define MAX_RECV_MSGS 50
 
-int32_t
-qb_ipcs_dispatch_service_request(int32_t fd, int32_t revents, void *data)
-{
-	int32_t res = _process_request_((struct qb_ipcs_connection *)data,
-					IPC_REQUEST_TIMEOUT);
-	if (res > 0) {
-		return 0;
-	}
-	return res;
-}
-
 static ssize_t
 _request_q_len_get(struct qb_ipcs_connection *c)
 {
@@ -703,22 +747,24 @@ qb_ipcs_dispatch_connection_request(int32_t fd, int32_t revents, void *data)
 {
 	struct qb_ipcs_connection *c = (struct qb_ipcs_connection *)data;
 	char bytes[MAX_RECV_MSGS];
-	int32_t res;
+	int32_t res = 0;
 	int32_t res2;
 	int32_t recvd = 0;
 	ssize_t avail;
 
 	if (revents & POLLNVAL) {
 		qb_util_log(LOG_DEBUG, "NVAL conn (%s)", c->description);
-		return -EINVAL;
+		res = -EINVAL;
+		goto dispatch_cleanup;
 	}
 	if (revents & POLLHUP) {
 		qb_util_log(LOG_DEBUG, "HUP conn (%s)", c->description);
-		qb_ipcs_disconnect(c);
-		return -ESHUTDOWN;
+		res = -ESHUTDOWN;
+		goto dispatch_cleanup;
 	}
 
 	if (revents & POLLOUT) {
+		/* try resend events now that fd can write */
 		res = resend_event_notifications(c);
 		if (res < 0 && res != -EAGAIN) {
 			errno = -res;
@@ -726,12 +772,15 @@ qb_ipcs_dispatch_connection_request(int32_t fd, int32_t revents, void *data)
 				       "resend_event_notifications (%s)",
 				       c->description);
 		}
+		/* nothing to read */
 		if ((revents & POLLIN) == 0) {
-			return 0;
+			res = 0;
+			goto dispatch_cleanup;
 		}
 	}
 	if (c->fc_enabled) {
-		return 0;
+		res = 0;
+		goto dispatch_cleanup;
 	}
 	avail = _request_q_len_get(c);
 
@@ -741,18 +790,24 @@ qb_ipcs_dispatch_connection_request(int32_t fd, int32_t revents, void *data)
 			errno = -res2;
 			qb_util_perror(LOG_WARNING, "conn (%s) disconnected",
 				       c->description);
-			qb_ipcs_disconnect(c);
-			return -ESHUTDOWN;
+			res = -ESHUTDOWN;
+			goto dispatch_cleanup;
 		} else {
 			qb_util_log(LOG_WARNING,
 				    "conn (%s) Nothing in q but got POLLIN on fd:%d (res2:%d)",
 				    c->description, fd, res2);
-			return 0;
+			res = 0;
+			goto dispatch_cleanup;
 		}
 	}
 
 	do {
 		res = _process_request_(c, IPC_REQUEST_TIMEOUT);
+
+		if (res == -ESHUTDOWN) {
+			goto dispatch_cleanup;
+		}
+
 		if (res > 0 || res == -ENOBUFS || res == -EINVAL) {
 			recvd++;
 		}
@@ -763,11 +818,12 @@ qb_ipcs_dispatch_connection_request(int32_t fd, int32_t revents, void *data)
 
 	if (c->service->needs_sock_for_poll && recvd > 0) {
 		res2 = qb_ipc_us_recv(&c->setup, bytes, recvd, -1);
-		if (res2 < 0) {
+		if (qb_ipc_us_sock_error_is_disconnected(res2)) {
 			errno = -res2;
-			qb_util_perror(LOG_ERR,
-				       "error receiving from setup sock (%s)",
-				       c->description);
+			qb_util_perror(LOG_ERR, "error receiving from setup sock (%s)", c->description);
+
+			res = -ESHUTDOWN;
+			goto dispatch_cleanup;
 		}
 	}
 
@@ -776,12 +832,20 @@ qb_ipcs_dispatch_connection_request(int32_t fd, int32_t revents, void *data)
 		res = 0;
 	}
 	if (res != 0) {
-		errno = -res;
-		qb_util_perror(LOG_ERR, "request returned error (%s)",
-			       c->description);
-		qb_ipcs_connection_unref(c);
+		if (res != -ENOTCONN) {
+			/*
+			 * Abnormal state (ENOTCONN is normal shutdown).
+			 */
+			errno = -res;
+			qb_util_perror(LOG_ERR, "request returned error (%s)",
+				       c->description);
+		}
 	}
 
+dispatch_cleanup:
+	if (res != 0) {
+		qb_ipcs_disconnect(c);
+	}
 	return res;
 }
 
@@ -803,6 +867,15 @@ qb_ipcs_context_get(struct qb_ipcs_connection *c)
 	return c->context;
 }
 
+void *
+qb_ipcs_connection_service_context_get(qb_ipcs_connection_t *c)
+{
+	if (c == NULL || c->service == NULL) {
+		return NULL;
+	}
+	return c->service->context;
+}
+
 int32_t
 qb_ipcs_connection_stats_get(qb_ipcs_connection_t * c,
 			     struct qb_ipcs_connection_stats * stats,
@@ -823,7 +896,7 @@ struct qb_ipcs_connection_stats_2*
 qb_ipcs_connection_stats_get_2(qb_ipcs_connection_t *c,
 			       int32_t clear_after_read)
 {
-	struct qb_ipcs_connection_stats_2* stats;
+	struct qb_ipcs_connection_stats_2 * stats;
 
 	if (c == NULL) {
 		errno = EINVAL;
@@ -872,3 +945,24 @@ qb_ipcs_connection_auth_set(qb_ipcs_connection_t *c, uid_t uid,
 		c->auth.mode = mode;
 	}
 }
+
+int32_t
+qb_ipcs_connection_get_buffer_size(qb_ipcs_connection_t *c)
+{
+	if (c == NULL) {
+		return -EINVAL;
+	}
+
+	/* request, response, and event shoud all have the same
+	 * buffer size allocated. It doesn't matter which we return
+	 * here. */
+	return c->response.max_msg_size;
+}
+
+void qb_ipcs_enforce_buffer_size(qb_ipcs_service_t *s, uint32_t buf_size)
+{
+	if (s == NULL) {
+		return;
+	}
+	s->max_buffer_size = buf_size;
+}
