@@ -37,6 +37,7 @@
 #include <qb/qbarray.h>
 #include "log_int.h"
 #include "util_int.h"
+#include <regex.h>
 
 static struct qb_log_target conf[QB_LOG_TARGET_MAX];
 static uint32_t conf_active_max = 0;
@@ -60,11 +61,13 @@ static void _log_filter_apply(struct callsite_section *sect,
 			      uint32_t t, enum qb_log_filter_conf c,
 			      enum qb_log_filter_type type,
 			      const char *text,
+			      regex_t *regex,
 			      uint8_t high_priority, uint8_t low_priority);
 static void _log_filter_apply_to_cs(struct qb_log_callsite *cs,
 				    uint32_t t, enum qb_log_filter_conf c,
 				    enum qb_log_filter_type type,
 				    const char *text,
+				    regex_t *regex,
 				    uint8_t high_priority, uint8_t low_priority);
 
 /* deprecated method of getting internal log messages */
@@ -75,14 +78,28 @@ qb_util_set_log_function(qb_util_log_fn_t fn)
 	old_internal_log_fn = fn;
 }
 
+static void
+_log_free_filter(struct qb_log_filter *flt)
+{
+	if (flt->regex) {
+		regfree(flt->regex);
+	}
+	free(flt->regex);
+	free(flt->text);
+	free(flt);
+}
+
 static int32_t
 _cs_matches_filter_(struct qb_log_callsite *cs,
 		    enum qb_log_filter_type type,
 		    const char *text,
+		    regex_t *regex,
 		    uint8_t high_priority,
 		    uint8_t low_priority)
 {
 	int32_t match = QB_FALSE;
+	const char *offset = NULL;
+	const char *next = NULL;
 
 	if (cs->priority > low_priority ||
 	    cs->priority < high_priority) {
@@ -91,32 +108,52 @@ _cs_matches_filter_(struct qb_log_callsite *cs,
 	if (strcmp(text, "*") == 0) {
 		return QB_TRUE;
 	}
-	if (type == QB_LOG_FILTER_FILE ||
-	    type == QB_LOG_FILTER_FUNCTION) {
-		char token[500];
-		const char *offset = NULL;
-		const char *next = text;
 
+	switch (type) {
+	case QB_LOG_FILTER_FILE:
+	case QB_LOG_FILTER_FUNCTION:
+		next = text;
 		do {
+			char token[500];
 			offset = next;
 			next = strchrnul(offset, ',');
 			snprintf(token, 499, "%.*s", (int)(next - offset), offset);
 
 			if (type == QB_LOG_FILTER_FILE) {
-				match = (strstr(cs->filename, token) != NULL);
+				match = (strcmp(cs->filename, token) == 0) ? 1 : 0;
 			} else {
-				match = (strstr(cs->function, token) != NULL);
+				match = (strcmp(cs->function, token) == 0) ? 1 : 0;
 			}
 			if (!match && next[0] != 0) {
 				next++;
 			}
 
 		} while (match == QB_FALSE && next != NULL && next[0] != 0);
-	} else if (type == QB_LOG_FILTER_FORMAT) {
+		break;
+	case QB_LOG_FILTER_FILE_REGEX:
+		next = next ? next : cs->filename;
+	case QB_LOG_FILTER_FUNCTION_REGEX:
+		next = next ? next : cs->function;
+	case QB_LOG_FILTER_FORMAT_REGEX:
+		next = next ? next : cs->format;
+
+		if (regex == NULL) {
+			return QB_FALSE;
+		}
+		match = regexec(regex, next, 0, NULL, 0);
+		if (match == 0) {
+			match = QB_TRUE;
+		} else {
+			match = QB_FALSE;
+		}
+		break;
+	case QB_LOG_FILTER_FORMAT:
 		if (strstr(cs->format, text)) {
 			match = QB_TRUE;
 		}
+		break;
 	}
+
 	return match;
 }
 
@@ -144,6 +181,8 @@ qb_log_real_va_(struct qb_log_callsite *cs, va_list ap)
 				va_copy(ap_copy, ap);
 				len = vsnprintf(str, QB_LOG_MAX_LEN, cs->format, ap_copy);
 				va_end(ap_copy);
+				if (len > QB_LOG_MAX_LEN)
+					len = QB_LOG_MAX_LEN;
 				if (str[len - 1] == '\n') str[len - 1] = '\0';
 				formatted = QB_TRUE;
 			}
@@ -173,6 +212,8 @@ qb_log_real_va_(struct qb_log_callsite *cs, va_list ap)
 					va_copy(ap_copy, ap);
 					len = vsnprintf(str, QB_LOG_MAX_LEN, cs->format, ap_copy);
 					va_end(ap_copy);
+					if (len > QB_LOG_MAX_LEN)
+						len = QB_LOG_MAX_LEN;
 					if (str[len - 1] == '\n') str[len - 1] = '\0';
 					formatted = QB_TRUE;
 				}
@@ -188,6 +229,8 @@ qb_log_real_va_(struct qb_log_callsite *cs, va_list ap)
 						va_copy(ap_copy, ap);
 						len = vsnprintf(str, QB_LOG_MAX_LEN, cs->format, ap_copy);
 						va_end(ap_copy);
+						if (len > QB_LOG_MAX_LEN)
+							len = QB_LOG_MAX_LEN;
 						if (str[len - 1] == '\n') str[len - 1] = '\0';
 						formatted = QB_TRUE;
 					}
@@ -266,18 +309,18 @@ qb_log_callsite_get(const char *function,
 			if (t->state != QB_LOG_STATE_ENABLED) {
 				continue;
 			}
-			for (f_item = t->filter_head.next; f_item != &t->filter_head; f_item = f_item->next) {
+			qb_list_for_each(f_item, &t->filter_head) {
 				flt = qb_list_entry(f_item, struct qb_log_filter, list);
 				_log_filter_apply_to_cs(cs, t->pos, flt->conf, flt->type,
-							flt->text, flt->high_priority,
+							flt->text, flt->regex, flt->high_priority,
 							flt->low_priority);
 			}
 		}
 		if (tags == 0) {
-			for (f_item = tags_head.next; f_item != &tags_head; f_item = f_item->next) {
+			qb_list_for_each(f_item, &tags_head) {
 				flt = qb_list_entry(f_item, struct qb_log_filter, list);
 				_log_filter_apply_to_cs(cs, flt->new_value, flt->conf, flt->type,
-							flt->text, flt->high_priority,
+							flt->text, flt->regex, flt->high_priority,
 							flt->low_priority);
 			}
 		} else {
@@ -379,13 +422,13 @@ qb_log_callsites_register(struct qb_log_callsite *_start,
 		}
 		qb_list_for_each_entry(flt, &t->filter_head, list) {
 			_log_filter_apply(sect, t->pos, flt->conf,
-					  flt->type, flt->text,
+					  flt->type, flt->text, flt->regex,
 					  flt->high_priority, flt->low_priority);
 		}
 	}
 	qb_list_for_each_entry(flt, &tags_head, list) {
 		_log_filter_apply(sect, flt->new_value, flt->conf,
-				  flt->type, flt->text,
+				  flt->type, flt->text, flt->regex,
 				  flt->high_priority, flt->low_priority);
 	}
 	pthread_rwlock_unlock(&_listlock);
@@ -459,7 +502,8 @@ _log_filter_store(uint32_t t, enum qb_log_filter_conf c,
 		  enum qb_log_filter_type type,
 		  const char *text,
 		  uint8_t high_priority,
-		  uint8_t low_priority)
+		  uint8_t low_priority,
+		  struct qb_log_filter **new)
 {
 	struct qb_log_filter *flt;
 	struct qb_list_head *iter;
@@ -499,13 +543,33 @@ _log_filter_store(uint32_t t, enum qb_log_filter_conf c,
 		flt->type = type;
 		flt->text = strdup(text);
 		if (flt->text == NULL) {
-			free(flt);
+			_log_free_filter(flt);
 			return -ENOMEM;
 		}
+
+		if (type == QB_LOG_FILTER_FUNCTION_REGEX ||
+			type == QB_LOG_FILTER_FILE_REGEX ||
+			type == QB_LOG_FILTER_FORMAT_REGEX) {
+			int res;
+
+			flt->regex = calloc(1, sizeof(regex_t));
+			if (flt->regex == NULL) {
+				_log_free_filter(flt);
+				return -ENOMEM;
+			}
+			res = regcomp(flt->regex, flt->text, 0);
+			if (res) {
+				_log_free_filter(flt);
+				return -EINVAL;
+			}
+		}
 		flt->high_priority = high_priority;
 		flt->low_priority = low_priority;
 		flt->new_value = t;
 		qb_list_add_tail(&flt->list, list_head);
+		if (new) {
+			*new = flt;
+		}
 	} else if (c == QB_LOG_FILTER_REMOVE || c == QB_LOG_TAG_CLEAR) {
 		qb_list_for_each_safe(iter, next, list_head) {
 			flt = qb_list_entry(iter, struct qb_log_filter, list);
@@ -515,8 +579,7 @@ _log_filter_store(uint32_t t, enum qb_log_filter_conf c,
 			    (strcmp(flt->text, text) == 0 ||
 			     strcmp("*", text) == 0)) {
 				qb_list_del(iter);
-				free(flt->text);
-				free(flt);
+				_log_free_filter(flt);
 				return 0;
 			}
 		}
@@ -525,8 +588,7 @@ _log_filter_store(uint32_t t, enum qb_log_filter_conf c,
 		qb_list_for_each_safe(iter, next, list_head) {
 			flt = qb_list_entry(iter, struct qb_log_filter, list);
 			qb_list_del(iter);
-			free(flt->text);
-			free(flt);
+			_log_free_filter(flt);
 		}
 	}
 	return 0;
@@ -537,13 +599,14 @@ _log_filter_apply(struct callsite_section *sect,
 		  uint32_t t, enum qb_log_filter_conf c,
 		  enum qb_log_filter_type type,
 		  const char *text,
+		  regex_t *regex,
 		  uint8_t high_priority, uint8_t low_priority)
 {
 	struct qb_log_callsite *cs;
 
 	for (cs = sect->start; cs < sect->stop; cs++) {
 		if (cs->lineno > 0) {
-			_log_filter_apply_to_cs(cs, t, c, type, text,
+			_log_filter_apply_to_cs(cs, t, c, type, text, regex,
 					    high_priority, low_priority);
 		}
 	}
@@ -555,6 +618,7 @@ _log_filter_apply_to_cs(struct qb_log_callsite *cs,
 			uint32_t t, enum qb_log_filter_conf c,
 			enum qb_log_filter_type type,
 			const char *text,
+			regex_t *regex,
 			uint8_t high_priority, uint8_t low_priority)
 {
 
@@ -566,7 +630,7 @@ _log_filter_apply_to_cs(struct qb_log_callsite *cs,
 		return;
 	}
 
-	if (_cs_matches_filter_(cs, type, text, high_priority, low_priority)) {
+	if (_cs_matches_filter_(cs, type, text, regex, high_priority, low_priority)) {
 #ifdef _QB_FILTER_DEBUGGING_
 		uint32_t old_targets = cs->targets;
 		uint32_t old_tags = cs->tags;
@@ -599,6 +663,8 @@ qb_log_filter_ctl2(int32_t t, enum qb_log_filter_conf c,
 		   enum qb_log_filter_type type, const char * text,
 		   uint8_t high_priority, uint8_t low_priority)
 {
+	struct qb_log_filter *new_flt = NULL;
+	regex_t *regex = NULL;
 	struct callsite_section *sect;
 	int32_t rc;
 
@@ -622,14 +688,17 @@ qb_log_filter_ctl2(int32_t t, enum qb_log_filter_conf c,
 		return -EINVAL;
 	}
 	pthread_rwlock_rdlock(&_listlock);
-	rc = _log_filter_store(t, c, type, text, high_priority, low_priority);
+	rc = _log_filter_store(t, c, type, text, high_priority, low_priority, &new_flt);
 	if (rc < 0) {
 		pthread_rwlock_unlock(&_listlock);
 		return rc;
 	}
 
+	if (new_flt && new_flt->regex) {
+		regex = new_flt->regex;
+	}
 	qb_list_for_each_entry(sect, &callsite_sections, list) {
-		_log_filter_apply(sect, t, c, type, text, high_priority, low_priority);
+		_log_filter_apply(sect, t, c, type, text, regex, high_priority, low_priority);
 	}
 	pthread_rwlock_unlock(&_listlock);
 	return 0;
@@ -735,6 +804,7 @@ qb_log_init(const char *name, int32_t facility, uint8_t priority)
 
 	i = pthread_rwlock_init(&_listlock, NULL);
 	assert(i == 0);
+	qb_log_format_init();
 
 	for (i = 0; i < QB_LOG_TARGET_MAX; i++) {
 		conf[i].pos = i;
@@ -744,7 +814,6 @@ qb_log_init(const char *name, int32_t facility, uint8_t priority)
 		(void)strlcpy(conf[i].name, name, PATH_MAX);
 		conf[i].facility = facility;
 		qb_list_init(&conf[i].filter_head);
-		qb_log_format_set(i, NULL);
 	}
 
 	qb_log_dcs_init();
@@ -786,14 +855,13 @@ qb_log_fini(void)
 	for (pos = 0; pos <= conf_active_max; pos++) {
 		t = &conf[pos];
 		_log_target_disable(t);
-		free(conf[pos].format);
 		qb_list_for_each_safe(iter2, next2, &t->filter_head) {
 			flt = qb_list_entry(iter2, struct qb_log_filter, list);
 			qb_list_del(iter2);
-			free(flt->text);
-			free(flt);
+			_log_free_filter(flt);
 		}
 	}
+	qb_log_format_fini();
 	qb_log_dcs_fini();
 	qb_list_for_each_safe(iter, next, &callsite_sections) {
 		s = qb_list_entry(iter, struct callsite_section, list);
@@ -803,8 +871,7 @@ qb_log_fini(void)
 	qb_list_for_each_safe(iter, next, &tags_head) {
 		flt = qb_list_entry(iter, struct qb_log_filter, list);
 		qb_list_del(iter);
-		free(flt->text);
-		free(flt);
+		_log_free_filter(flt);
 	}
 }
 
@@ -1008,18 +1075,3 @@ qb_log_ctl(int32_t t, enum qb_log_conf c, int32_t arg)
 	}
 	return rc;
 }
-
-void
-qb_log_format_set(int32_t t, const char *format)
-{
-	char modified_format[256];
-	free(conf[t].format);
-
-	if (format) {
-		qb_log_target_format_static(t, format, modified_format);
-		conf[t].format = strdup(modified_format);
-	} else {
-		conf[t].format = strdup("[%p] %b");
-	}
-	assert(conf[t].format != NULL);
-}
