@@ -24,6 +24,7 @@
 #include <qb/qblist.h>
 #include <qb/qblog.h>
 #include <qb/qbrb.h>
+#include <regex.h>
 
 struct qb_log_target;
 
@@ -56,6 +57,7 @@ struct qb_log_filter {
 	uint8_t low_priority;
 	uint32_t new_value;
 	struct qb_list_head list;
+	regex_t *regex;
 };
 
 struct qb_log_record {
@@ -94,6 +96,8 @@ struct qb_log_callsite *qb_log_dcs_get(int32_t *newly_created,
 				       uint32_t lineno,
 				       uint32_t tags);
 
+void qb_log_format_init(void);
+void qb_log_format_fini(void);
 const char * qb_log_priority2str(uint8_t priority);
 size_t qb_vsnprintf_serialize(char *serialize, size_t max_len, const char *fmt, va_list ap);
 size_t qb_vsnprintf_deserialize(char *string, size_t str_len, const char *buf);
