@@ -33,7 +33,22 @@
 #include <qb/qbloop.h>
 
 static const char *ipc_name = "ipc_test";
-#define MAX_MSG_SIZE (8192*16)
+
+#define DEFAULT_MAX_MSG_SIZE (8192*16)
+static int CALCULATED_DGRAM_MAX_MSG_SIZE = 0;
+
+#define DGRAM_MAX_MSG_SIZE \
+	(CALCULATED_DGRAM_MAX_MSG_SIZE == 0 ? \
+	CALCULATED_DGRAM_MAX_MSG_SIZE = qb_ipcc_verify_dgram_max_msg_size(DEFAULT_MAX_MSG_SIZE) : \
+	CALCULATED_DGRAM_MAX_MSG_SIZE)
+
+#define MAX_MSG_SIZE (ipc_type == QB_IPC_SOCKET ? DGRAM_MAX_MSG_SIZE : DEFAULT_MAX_MSG_SIZE)
+
+/* The size the giant msg's data field needs to be to make
+ * this the largests msg we can successfully send. */
+#define GIANT_MSG_DATA_SIZE MAX_MSG_SIZE - sizeof(struct qb_ipc_response_header) - 8
+
+static int enforce_server_buffer=0;
 static qb_ipcc_connection_t *conn;
 static enum qb_ipc_type ipc_type;
 
@@ -44,6 +59,8 @@ enum my_msg_ids {
 	IPC_MSG_RES_DISPATCH,
 	IPC_MSG_REQ_BULK_EVENTS,
 	IPC_MSG_RES_BULK_EVENTS,
+	IPC_MSG_REQ_STRESS_EVENT,
+	IPC_MSG_RES_STRESS_EVENT,
 	IPC_MSG_REQ_SERVER_FAIL,
 	IPC_MSG_RES_SERVER_FAIL,
 	IPC_MSG_REQ_SERVER_DISCONNECT,
@@ -75,6 +92,9 @@ static int32_t fc_enabled = 89;
 static int32_t send_event_on_created = QB_FALSE;
 static int32_t disconnect_after_created = QB_FALSE;
 static int32_t num_bulk_events = 10;
+static int32_t num_stress_events = 30000;
+static int32_t reference_count_test = QB_FALSE;
+
 
 static int32_t
 exit_handler(int32_t rsignal, void *data)
@@ -89,7 +109,7 @@ s1_msg_process_fn(qb_ipcs_connection_t *c,
 		void *data, size_t size)
 {
 	struct qb_ipc_request_header *req_pt = (struct qb_ipc_request_header *)data;
-	struct qb_ipc_response_header response;
+	struct qb_ipc_response_header response = { 0, };
 	ssize_t res;
 
 	if (req_pt->id == IPC_MSG_REQ_TX_RX) {
@@ -127,9 +147,31 @@ s1_msg_process_fn(qb_ipcs_connection_t *c,
 		num = stats->event_q_length;
 		free(stats);
 
-		for (m = 0; m < num_bulk_events; m++) {
-			res = qb_ipcs_event_send(c, &response,
-						 sizeof(response));
+		/* crazy large message */
+		res = qb_ipcs_event_send(c, &response,
+					 MAX_MSG_SIZE*10);
+		ck_assert_int_eq(res, -EMSGSIZE);
+
+		/* send one event before responding */
+		res = qb_ipcs_event_send(c, &response, sizeof(response));
+		ck_assert_int_eq(res, sizeof(response));
+		response.id++;
+
+		/* send response */
+		response.id = IPC_MSG_RES_BULK_EVENTS;
+		res = qb_ipcs_response_send(c, &response, response.size);
+		ck_assert_int_eq(res, sizeof(response));
+
+		/* send the rest of the events after the response */
+		for (m = 1; m < num_bulk_events; m++) {
+			res = qb_ipcs_event_send(c, &response, sizeof(response));
+
+			if (res == -EAGAIN || res == -ENOBUFS) {
+				/* retry */
+				usleep(1000);
+				m--;
+				continue;
+			}
 			ck_assert_int_eq(res, sizeof(response));
 			response.id++;
 		}
@@ -137,10 +179,49 @@ s1_msg_process_fn(qb_ipcs_connection_t *c,
 		ck_assert_int_eq(stats->event_q_length - num, num_bulk_events);
 		free(stats);
 
-		response.id = IPC_MSG_RES_BULK_EVENTS;
+	} else if (req_pt->id == IPC_MSG_REQ_STRESS_EVENT) {
+		struct {
+			struct qb_ipc_response_header hdr __attribute__ ((aligned(8)));
+			char data[GIANT_MSG_DATA_SIZE] __attribute__ ((aligned(8)));
+			uint32_t sent_msgs __attribute__ ((aligned(8)));
+		} __attribute__ ((aligned(8))) giant_event_send;
+		int32_t m;
+
+		response.size = sizeof(struct qb_ipc_response_header);
+		response.error = 0;
+
+		response.id = IPC_MSG_RES_STRESS_EVENT;
 		res = qb_ipcs_response_send(c, &response, response.size);
 		ck_assert_int_eq(res, sizeof(response));
 
+		giant_event_send.hdr.error = 0;
+		giant_event_send.hdr.id = IPC_MSG_RES_STRESS_EVENT;
+		for (m = 0; m < num_stress_events; m++) {
+			size_t sent_len = sizeof(struct qb_ipc_response_header);
+
+			if (((m+1) % 1000) == 0) {
+				sent_len = sizeof(giant_event_send);
+				giant_event_send.sent_msgs = m + 1;
+			}
+			giant_event_send.hdr.size = sent_len;
+
+			res = qb_ipcs_event_send(c, &giant_event_send, sent_len);
+			if (res < 0) {
+				if (res == -EAGAIN || res == -ENOBUFS) {
+					/* yield to the receive process */
+					usleep(1000);
+					m--;
+					continue;
+				} else {
+					qb_perror(LOG_DEBUG, "sending stress events");
+					ck_assert_int_eq(res, sent_len);
+				}
+			} else if (((m+1) % 1000) == 0) {
+				qb_log(LOG_DEBUG, "SENT: %d stress events sent", m+1);
+			}
+			giant_event_send.hdr.id++;
+		}
+
 	} else if (req_pt->id == IPC_MSG_REQ_SERVER_FAIL) {
 		exit(0);
 	} else if (req_pt->id == IPC_MSG_REQ_SERVER_DISCONNECT) {
@@ -185,17 +266,55 @@ s1_connection_closed(qb_ipcs_connection_t *c)
 	return 0;
 }
 
+static void
+outq_flush (void *data)
+{
+	static int i = 0;
+	struct cs_ipcs_conn_context *cnx;
+	cnx = qb_ipcs_context_get(data);
+
+	qb_log(LOG_DEBUG,"iter %u\n", i);
+	i++;
+	if (i == 2) {
+		qb_ipcs_destroy(s1);
+		s1 = NULL;
+	}
+	/* is the reference counting is not working, this should fail
+	 * for i > 1.
+	 */
+	qb_ipcs_event_send(data, "test", 4);
+	assert(memcmp(cnx, "test", 4) == 0);
+	if (i < 5) {
+		qb_loop_job_add(my_loop, QB_LOOP_HIGH, data, outq_flush);
+	} else {
+		/* this single unref should clean everything up.
+		 */
+		qb_ipcs_connection_unref(data);
+		qb_log(LOG_INFO, "end of test, stopping loop");
+		qb_loop_stop(my_loop);
+	}
+}
+
+
 static void
 s1_connection_destroyed(qb_ipcs_connection_t *c)
 {
 	qb_enter();
-	qb_loop_stop(my_loop);
+	if (reference_count_test) {
+		struct cs_ipcs_conn_context *cnx;
+		cnx = qb_ipcs_context_get(c);
+		free(cnx);
+	} else {
+		qb_loop_stop(my_loop);
+	}
 	qb_leave();
 }
 
 static void
 s1_connection_created(qb_ipcs_connection_t *c)
 {
+	int32_t max = MAX_MSG_SIZE;
+
 	if (send_event_on_created) {
 		struct qb_ipc_response_header response;
 		int32_t res;
@@ -207,6 +326,20 @@ s1_connection_created(qb_ipcs_connection_t *c)
 					 sizeof(response));
 		ck_assert_int_eq(res, response.size);
 	}
+	if (reference_count_test) {
+		struct cs_ipcs_conn_context *context;
+
+		qb_ipcs_connection_ref(c);
+		qb_loop_job_add(my_loop, QB_LOOP_HIGH, c, outq_flush);
+
+		context = calloc(1, 20);
+		memcpy(context, "test", 4);
+		qb_ipcs_context_set(c, context);
+	}
+
+
+	ck_assert_int_eq(max, qb_ipcs_connection_get_buffer_size(c));
+
 }
 
 static void
@@ -240,12 +373,16 @@ run_ipc_server(void)
 	s1 = qb_ipcs_create(ipc_name, 4, ipc_type, &sh);
 	fail_if(s1 == 0);
 
+	if (enforce_server_buffer) {
+		qb_ipcs_enforce_buffer_size(s1, MAX_MSG_SIZE);
+	}
 	qb_ipcs_poll_handlers_set(s1, &ph);
 
 	res = qb_ipcs_run(s1);
 	ck_assert_int_eq(res, 0);
 
 	qb_loop_run(my_loop);
+	qb_log(LOG_DEBUG, "loop finished - done ...");
 }
 
 static int32_t
@@ -292,8 +429,13 @@ send_and_check(int32_t req_id, uint32_t size,
 	request.hdr.id = req_id;
 	request.hdr.size = sizeof(struct qb_ipc_request_header) + size;
 
-repeat_send:
+	/* check that we can't send a message that is too big
+	 * and we get the right return code.
+	 */
+	res = qb_ipcc_send(conn, &request, MAX_MSG_SIZE*2);
+	ck_assert_int_eq(res, -EMSGSIZE);
 
+repeat_send:
 	res = qb_ipcc_send(conn, &request, request.hdr.size);
 	try_times++;
 	if (res < 0) {
@@ -575,13 +717,59 @@ END_TEST
 
 static int32_t events_received;
 
+static int32_t
+count_stress_events(int32_t fd, int32_t revents, void *data)
+{
+	struct {
+		struct qb_ipc_response_header hdr __attribute__ ((aligned(8)));
+		char data[GIANT_MSG_DATA_SIZE] __attribute__ ((aligned(8)));
+		uint32_t sent_msgs __attribute__ ((aligned(8)));
+	} __attribute__ ((aligned(8))) giant_event_recv;
+	qb_loop_t *cl = (qb_loop_t*)data;
+	int32_t res;
+
+	res = qb_ipcc_event_recv(conn, &giant_event_recv,
+				 sizeof(giant_event_recv),
+				 -1);
+	if (res > 0) {
+		events_received++;
+
+		if ((events_received % 1000) == 0) {
+			qb_log(LOG_DEBUG, "RECV: %d stress events processed", events_received);
+			if (res != sizeof(giant_event_recv)) {
+				qb_log(LOG_DEBUG, "Unexpected recv size, expected %d got %d",
+					res, sizeof(giant_event_recv));
+			} else if (giant_event_recv.sent_msgs != events_received) {
+				qb_log(LOG_DEBUG, "Server event mismatch. Server thinks we got %d msgs, but we only received %d",
+					giant_event_recv.sent_msgs, events_received);
+			}
+		}
+	} else if (res != -EAGAIN) {
+		qb_perror(LOG_DEBUG, "count_stress_events");
+		qb_loop_stop(cl);
+		return -1;
+	}
+
+	if (events_received >= num_stress_events) {
+		qb_loop_stop(cl);
+		return -1;
+	}
+	return 0;
+}
 
 static int32_t
 count_bulk_events(int32_t fd, int32_t revents, void *data)
 {
 	qb_loop_t *cl = (qb_loop_t*)data;
+	struct qb_ipc_response_header res_header;
+	int32_t res;
 
-	events_received++;
+	res = qb_ipcc_event_recv(conn, &res_header,
+				 sizeof(struct qb_ipc_response_header),
+				 -1);
+	if (res > 0) {
+		events_received++;
+	}
 
 	if (events_received >= num_bulk_events) {
 		qb_loop_stop(cl);
@@ -620,7 +808,7 @@ test_ipc_bulk_events(void)
 
 	events_received = 0;
 	cl = qb_loop_create();
-	res = qb_ipcc_fd_get(conn, &fd),
+	res = qb_ipcc_fd_get(conn, &fd);
 	ck_assert_int_eq(res, 0);
 	res = qb_loop_poll_add(cl, QB_LOOP_MED,
 			 fd, POLLIN,
@@ -652,6 +840,100 @@ test_ipc_bulk_events(void)
 	stop_process(pid);
 }
 
+static void
+test_ipc_stress_test(void)
+{
+	struct {
+		struct qb_ipc_request_header hdr __attribute__ ((aligned(8)));
+		char data[GIANT_MSG_DATA_SIZE] __attribute__ ((aligned(8)));
+		uint32_t sent_msgs __attribute__ ((aligned(8)));
+	} __attribute__ ((aligned(8))) giant_req;
+
+	struct qb_ipc_response_header res_header;
+	struct iovec iov[1];
+	int32_t c = 0;
+	int32_t j = 0;
+	pid_t pid;
+	int32_t res;
+	qb_loop_t *cl;
+	int32_t fd;
+	/* This looks strange, but it serves an important purpose.
+	 * This test forces the server to enforce the MAX_MSG_SIZE
+	 * limit from the server side, which overrides the client's
+	 * buffer limit.  To verify this functionality is working
+	 * we set the client limit lower than what the server
+	 * is enforcing. */
+	int32_t client_buf_size = MAX_MSG_SIZE - 1024;
+	int32_t real_buf_size;
+
+	enforce_server_buffer = 1;
+	pid = run_function_in_new_process(run_ipc_server);
+	enforce_server_buffer = 0;
+	fail_if(pid == -1);
+	sleep(1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, client_buf_size);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			sleep(1);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	real_buf_size = qb_ipcc_get_buffer_size(conn);
+	ck_assert_int_eq(real_buf_size, MAX_MSG_SIZE);
+
+	qb_log(LOG_DEBUG, "Testing %d iterations of EVENT msg passing.", num_stress_events);
+
+	events_received = 0;
+	cl = qb_loop_create();
+	res = qb_ipcc_fd_get(conn, &fd);
+	ck_assert_int_eq(res, 0);
+	res = qb_loop_poll_add(cl, QB_LOOP_MED,
+			 fd, POLLIN,
+			 cl, count_stress_events);
+	ck_assert_int_eq(res, 0);
+
+	res = send_and_check(IPC_MSG_REQ_STRESS_EVENT, 0, recv_timeout, QB_TRUE);
+
+	qb_loop_run(cl);
+	ck_assert_int_eq(events_received, num_stress_events);
+
+	giant_req.hdr.id = IPC_MSG_REQ_SERVER_FAIL;
+	giant_req.hdr.size = sizeof(giant_req);
+
+	if (giant_req.hdr.size <= client_buf_size) {
+		ck_assert_int_eq(1, 0);
+	}
+
+	iov[0].iov_len = giant_req.hdr.size;
+	iov[0].iov_base = &giant_req;
+	res = qb_ipcc_sendv_recv(conn, iov, 1,
+				 &res_header,
+				 sizeof(struct qb_ipc_response_header), -1);
+	if (res != -ECONNRESET && res != -ENOTCONN) {
+		qb_log(LOG_ERR, "id:%d size:%d", res_header.id, res_header.size);
+		ck_assert_int_eq(res, -ENOTCONN);
+	}
+
+	qb_ipcc_disconnect(conn);
+	stop_process(pid);
+}
+
+START_TEST(test_ipc_stress_test_us)
+{
+	qb_enter();
+	send_event_on_created = QB_FALSE;
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	test_ipc_stress_test();
+	qb_leave();
+}
+END_TEST
+
 START_TEST(test_ipc_bulk_events_us)
 {
 	qb_enter();
@@ -692,7 +974,7 @@ test_ipc_event_on_created(void)
 
 	events_received = 0;
 	cl = qb_loop_create();
-	res = qb_ipcc_fd_get(conn, &fd),
+	res = qb_ipcc_fd_get(conn, &fd);
 	ck_assert_int_eq(res, 0);
 	res = qb_loop_poll_add(cl, QB_LOOP_MED,
 			 fd, POLLIN,
@@ -755,9 +1037,12 @@ test_ipc_disconnect_after_created(void)
 				 &res_header,
 				 sizeof(struct qb_ipc_response_header), -1);
 	/*
-	 * confirm we get -ENOTCONN
+	 * confirm we get -ENOTCONN or -ECONNRESET
 	 */
-	ck_assert_int_eq(res, -ENOTCONN);
+	if (res != -ECONNRESET && res != -ENOTCONN) {
+		qb_log(LOG_ERR, "id:%d size:%d", res_header.id, res_header.size);
+		ck_assert_int_eq(res, -ENOTCONN);
+	}
 	ck_assert_int_eq(QB_FALSE, qb_ipcc_is_connected(conn));
 
 	qb_ipcc_disconnect(conn);
@@ -816,9 +1101,12 @@ test_ipc_server_fail(void)
 				 &res_header,
 				 sizeof(struct qb_ipc_response_header), -1);
 	/*
-	 * confirm we get -ENOTCONN
+	 * confirm we get -ENOTCONN or ECONNRESET
 	 */
-	ck_assert_int_eq(res, -ENOTCONN);
+	if (res != -ECONNRESET && res != -ENOTCONN) {
+		qb_log(LOG_ERR, "id:%d size:%d", res_header.id, res_header.size);
+		ck_assert_int_eq(res, -ENOTCONN);
+	}
 	ck_assert_int_eq(QB_FALSE, qb_ipcc_is_connected(conn));
 
 	qb_ipcc_disconnect(conn);
@@ -845,6 +1133,17 @@ START_TEST(test_ipc_disp_shm)
 }
 END_TEST
 
+START_TEST(test_ipc_stress_test_shm)
+{
+	qb_enter();
+	send_event_on_created = QB_FALSE;
+	ipc_type = QB_IPC_SHM;
+	ipc_name = __func__;
+	test_ipc_stress_test();
+	qb_leave();
+}
+END_TEST
+
 START_TEST(test_ipc_bulk_events_shm)
 {
 	qb_enter();
@@ -876,6 +1175,87 @@ START_TEST(test_ipc_server_fail_shm)
 }
 END_TEST
 
+static void
+test_ipc_service_ref_count(void)
+{
+	int32_t c = 0;
+	int32_t j = 0;
+	pid_t pid;
+
+	reference_count_test = QB_TRUE;
+
+	pid = run_function_in_new_process(run_ipc_server);
+	fail_if(pid == -1);
+	sleep(1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, MAX_MSG_SIZE);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			sleep(1);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	sleep(5);
+
+	qb_ipcc_disconnect(conn);
+	stop_process(pid);
+}
+
+
+START_TEST(test_ipc_service_ref_count_shm)
+{
+	qb_enter();
+	ipc_type = QB_IPC_SHM;
+	ipc_name = __func__;
+	test_ipc_service_ref_count();
+	qb_leave();
+}
+END_TEST
+
+START_TEST(test_ipc_service_ref_count_us)
+{
+	qb_enter();
+	ipc_type = QB_IPC_SOCKET;
+	ipc_name = __func__;
+	test_ipc_service_ref_count();
+	qb_leave();
+}
+END_TEST
+
+static void test_max_dgram_size(void)
+{
+	/* most implementations will not let you set a dgram buffer 
+	 * of 1 million bytes. This test verifies that the we can detect
+	 * the max dgram buffersize regardless, and that the value we detect
+	 * is consistent. */
+	int32_t init;
+	int32_t i;
+
+	qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_REMOVE,
+			  QB_LOG_FILTER_FILE, "*", LOG_TRACE);
+
+	init = qb_ipcc_verify_dgram_max_msg_size(1000000);
+	fail_if(init <= 0);
+	for (i = 0; i < 100; i++) {
+		int try = qb_ipcc_verify_dgram_max_msg_size(1000000);
+		ck_assert_int_eq(init, try);
+	}
+
+	qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD,
+			  QB_LOG_FILTER_FILE, "*", LOG_TRACE);
+}
+
+START_TEST(test_ipc_max_dgram_size)
+{
+	qb_enter();
+	test_max_dgram_size();
+	qb_leave();
+}
+END_TEST
 
 static Suite *
 make_shm_suite(void)
@@ -908,6 +1288,11 @@ make_shm_suite(void)
 	tcase_set_timeout(tc, 16);
 	suite_add_tcase(s, tc);
 
+	tc = tcase_create("ipc_stress_test_shm");
+	tcase_add_test(tc, test_ipc_stress_test_shm);
+	tcase_set_timeout(tc, 16);
+	suite_add_tcase(s, tc);
+
 	tc = tcase_create("ipc_bulk_events_shm");
 	tcase_add_test(tc, test_ipc_bulk_events_shm);
 	tcase_set_timeout(tc, 16);
@@ -922,6 +1307,11 @@ make_shm_suite(void)
 	tcase_add_test(tc, test_ipc_event_on_created_shm);
 	suite_add_tcase(s, tc);
 
+	tc = tcase_create("ipc_service_ref_count_shm");
+	tcase_add_test(tc, test_ipc_service_ref_count_shm);
+	tcase_set_timeout(tc, 10);
+	suite_add_tcase(s, tc);
+
 	return s;
 }
 
@@ -931,6 +1321,11 @@ make_soc_suite(void)
 	Suite *s = suite_create("socket");
 	TCase *tc;
 
+	tc = tcase_create("ipc_max_dgram_size");
+	tcase_add_test(tc, test_ipc_max_dgram_size);
+	tcase_set_timeout(tc, 30);
+	suite_add_tcase(s, tc);
+
 	tc = tcase_create("ipc_server_fail_soc");
 	tcase_add_test(tc, test_ipc_server_fail_soc);
 	tcase_set_timeout(tc, 6);
@@ -961,6 +1356,11 @@ make_soc_suite(void)
 	tcase_set_timeout(tc, 16);
 	suite_add_tcase(s, tc);
 
+	tc = tcase_create("ipc_stress_test_us");
+	tcase_add_test(tc, test_ipc_stress_test_us);
+	tcase_set_timeout(tc, 60);
+	suite_add_tcase(s, tc);
+
 	tc = tcase_create("ipc_bulk_events_us");
 	tcase_add_test(tc, test_ipc_bulk_events_us);
 	tcase_set_timeout(tc, 16);
@@ -974,6 +1374,11 @@ make_soc_suite(void)
 	tcase_add_test(tc, test_ipc_disconnect_after_created_us);
 	suite_add_tcase(s, tc);
 
+	tc = tcase_create("ipc_service_ref_count_us");
+	tcase_add_test(tc, test_ipc_service_ref_count_us);
+	tcase_set_timeout(tc, 10);
+	suite_add_tcase(s, tc);
+
 	return s;
 }
 
@@ -1008,4 +1413,3 @@ main(void)
 	srunner_free(sr);
 	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
-
