@@ -103,6 +103,19 @@ START_TEST(test_va_serialize)
 
 	format_this_up_to(buf, 11, "123456789____");
 	ck_assert_str_eq(buf, "123456789_");
+
+	format_this(buf, "Client %s.%.9s wants to fence (%s) '%s' with device '%s'",
+		    "bla", "foooooooooooooooooo",
+		    "action", "target", "hoop");
+
+	ck_assert_str_eq(buf,
+			 "Client bla.foooooooo wants to fence (action) 'target' with device 'hoop'");
+
+	format_this(buf, "Node %s now has process list: %.32x (was %.32x)",
+		    "18builder", 2, 0);
+	ck_assert_str_eq(buf, "Node 18builder now has process list: 00000000000000000000000000000002 (was 00000000000000000000000000000000)");
+
+
 }
 END_TEST
 
@@ -178,6 +191,7 @@ _test_logger(int32_t t,
 	test_buf[0] = '\0';
 	qb_log_target_format(t, cs, timestamp, msg, test_buf);
 	test_priority = cs->priority;
+
 	num_msgs++;
 }
 
@@ -269,6 +283,7 @@ START_TEST(test_log_basic)
 	rc = qb_log_ctl(t, QB_LOG_CONF_ENABLED, QB_TRUE);
 	ck_assert_int_eq(rc, 0);
 
+	/* captures last log */
 	memset(test_buf, 0, sizeof(test_buf));
 	test_priority = 0;
 	num_msgs = 0;
@@ -285,6 +300,54 @@ START_TEST(test_log_basic)
 	ck_assert_int_eq(num_msgs, 1);
 	ck_assert_str_eq(test_buf, "Hello Angus, how are you?");
 
+
+	/* 
+	 * test filtering by file regex
+ 	 */
+	qb_log_filter_ctl(t, QB_LOG_FILTER_CLEAR_ALL,
+			  QB_LOG_FILTER_FORMAT, "*", LOG_TRACE);
+	qb_log_filter_ctl(t, QB_LOG_FILTER_ADD,
+			  QB_LOG_FILTER_FILE_REGEX, "^fakefile*", LOG_DEBUG);
+
+	num_msgs = 0;
+	qb_log_from_external_source(__func__, "fakefile_logging", "%s bla", LOG_INFO,
+				    56, 0, "filename/lineno");
+	qb_log_from_external_source(__func__, "do_not_log_fakefile_logging", "%s bla", LOG_INFO,
+				    56, 0, "filename/lineno");
+	ck_assert_int_eq(num_msgs, 1);
+
+	/* 
+	 * test filtering by format regex
+ 	 */
+	qb_log_filter_ctl(t, QB_LOG_FILTER_CLEAR_ALL,
+			  QB_LOG_FILTER_FORMAT, "*", LOG_TRACE);
+	qb_log_filter_ctl(t, QB_LOG_FILTER_ADD,
+			  QB_LOG_FILTER_FORMAT_REGEX, "^one", LOG_WARNING);
+
+	num_msgs = 0;
+	qb_log(LOG_INFO, "one two three");
+	qb_log(LOG_ERR, "testing one two three");
+	qb_log(LOG_WARNING, "one two three");
+	qb_log(LOG_ERR, "one two three");
+	qb_log(LOG_EMERG, "one two three");
+	ck_assert_int_eq(num_msgs, 3);
+
+	/*
+	 * test filtering by function and regex
+	 */
+	qb_log_filter_ctl(t, QB_LOG_FILTER_CLEAR_ALL,
+			  QB_LOG_FILTER_FILE, "*", LOG_TRACE);
+	qb_log_filter_ctl(t, QB_LOG_FILTER_ADD,
+			  QB_LOG_FILTER_FUNCTION_REGEX, "^log_.*please", LOG_WARNING);
+
+	num_msgs = 0;
+	qb_log(LOG_ERR, "try if you: log_it_please()");
+	log_it_please();
+	ck_assert_int_eq(num_msgs, 3);
+
+	qb_log_filter_ctl(t, QB_LOG_FILTER_REMOVE,
+			  QB_LOG_FILTER_FUNCTION_REGEX, "log_it_please", LOG_WARNING);
+
 	/*
 	 * test filtering by function
 	 */
@@ -323,7 +386,7 @@ START_TEST(test_log_basic)
 	qb_log_filter_ctl(t, QB_LOG_FILTER_CLEAR_ALL,
 			  QB_LOG_FILTER_FILE, "*", LOG_TRACE);
 	qb_log_filter_ctl(t, QB_LOG_FILTER_ADD,
-			  QB_LOG_FILTER_FILE, __FILE__, LOG_DEBUG);
+			  QB_LOG_FILTER_FILE, "fakefile.c,"__FILE__",otherfakefile", LOG_DEBUG);
 	/*
 	 * make sure we can pass in a null filename or function name.
 	 */
